from __future__ import annotations

import asyncio
import logging
import os
from abc import ABC, abstractmethod
from typing import cast, override

from openai import AsyncOpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCall,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolParam,
)

from relace_agent.config import AgentConfig, AgentConfigOverrides
from relace_agent.context import (
    AssistantMessage,
    Context,
    ToolUse,
    UserMessage,
)
from relace_agent.errors import AgentError, AgentStop
from relace_agent.server.types import (
    AgentEvent,
    ToolEvent,
)
from relace_toolbox.errors import ExitAgentLoop, ToolError
from relace_toolbox.tools import ToolSchema
from relace_toolbox.types import ToolResult

logger = logging.getLogger(__name__)


class Agent:
    def __init__(
        self,
        config: AgentConfig,
        config_overrides: AgentConfigOverrides | None = None,
    ) -> None:
        if config_overrides is not None:
            logger.info("Applying config overrides: %s", config_overrides)
            config = config.with_overrides(config_overrides)

        self.config: AgentConfig = config
        self.inference_provider: InferenceProvider = RelaceInferenceProvider(config)

    async def format_tool_definitions(
        self, context: Context
    ) -> list[ChatCompletionToolParam]:
        tool_specs = await context.toolbox.tool_specs()
        return [
            tool_spec.tool_schema
            for tool_spec in tool_specs.values()
            if tool_spec.tool_name in self.config.tools
        ]

    def format_system_prompt(
        self, context: Context
    ) -> ChatCompletionSystemMessageParam:
        return {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": context.interpolate(self.config.prompts.system_prompt),
                }
            ],
        }

    def format_history(self, context: Context) -> list[ChatCompletionMessageParam]:
        return context.history.as_params()

    async def generate_response(self, context: Context) -> ChatCompletion:
        return await self.inference_provider.generate_response(
            context=context,
            config=self.config,
            messages=[
                self.format_system_prompt(context),
                *self.format_history(context),
            ],
            tools=await self.format_tool_definitions(context),
        )

    async def handle_response(
        self,
        context: Context,
        response: ChatCompletion,
    ) -> None:
        logger.info("Agent response: %s", response)
        context.history.record_usage(response.usage)

        # Configured for one choice only
        choice = response.choices[0]
        if choice.message.tool_calls:
            logger.info("Agent performing tool calls")
            if choice.message.content:
                await context.events.emit(
                    AgentEvent(
                        name=self.config.name,
                        content=choice.message.content,
                    )
                )
            # Execute all tool calls concurrently
            tool_call_tasks = [
                self.handle_tool_call(
                    context=context,
                    tool_call=cast(ChatCompletionMessageToolCall, tool_call),
                )
                for tool_call in choice.message.tool_calls
            ]
            # NOTE: AgentStop may cause cancellation of other tool calls
            results = await asyncio.gather(*tool_call_tasks)
            tool_results = {
                tool_call.id: result
                for tool_call, result in zip(
                    choice.message.tool_calls, results, strict=True
                )
            }
            context.history.append(
                ToolUse(message=choice.message, results=tool_results)
            )
        elif choice.finish_reason == "stop":
            logger.info("Agent stopped naturally: %s", choice.finish_reason)
            if choice.message.content:
                await context.events.emit(
                    AgentEvent(name=self.config.name, content=choice.message.content)
                )
                context.history.append(AssistantMessage(content=choice.message.content))
            raise AgentStop(choice.finish_reason)
        else:
            logger.error("Agent stopped unexpectedly: %s", choice.finish_reason)
            raise AgentError(f"Agent stopped unexpectedly: {choice.finish_reason}")

    async def handle_tool_call(
        self,
        context: Context,
        tool_call: ChatCompletionMessageToolCall,
    ) -> ToolResult:
        try:
            tool_schema = ToolSchema.from_name(tool_call.function.name)
            tool_input = tool_schema.model_validate_json(tool_call.function.arguments)
            tool_event = tool_input.event_data()
        except Exception as exc:
            # NOTE: This may happen due to a mismatch between the schema version in
            # the agent runtime and toolbox. We try to process the tool call in this
            # case - if the input is actually malformed, we will get a ToolError below.
            # We may want to delegate event data construction to the toolbox later to
            # avoid this issue.
            logger.warning("Failed to parse tool event: %s", exc)
            tool_event = {}

        try:
            await context.events.emit(
                ToolEvent(
                    name=tool_call.function.name,
                    **tool_event,
                )
            )
            tool_spec = await context.toolbox.tool_spec(tool_call.function.name)
            return await context.toolbox.tool_call(
                tool_name=tool_call.function.name,
                tool_input=tool_call.function.arguments,
                workdir=context.toolbox_workdir,
                chat_history=context.history.tokenize()
                if tool_spec.requires_chat_history
                else None,
            )
        except ExitAgentLoop as exit_signal:
            raise AgentStop("Tool emitted exit signal") from exit_signal
        except ToolError as error:
            logger.warning("Tool error: %s", error)
            return ToolResult(message=f"Error: {error}")

    async def run_loop(self, context: Context) -> None:
        """Run the agent with the given query."""
        user_prompt = self.config.prompts.user_prompt

        context.history.append(UserMessage(content=context.interpolate(user_prompt)))
        turn = 0
        while turn < self.config.max_turns:
            turn += 1
            response = await self.generate_response(context)

            try:
                await self.handle_response(context, response)
            except AgentStop:
                break
        else:
            logger.warning("Agent reached turn limit: %s", self.config.max_turns)


class InferenceProvider(ABC):
    @abstractmethod
    async def generate_response(
        self,
        context: Context,
        config: AgentConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam],
    ) -> ChatCompletion:
        """Generate a response from the model."""
        ...


class RelaceInferenceProvider(InferenceProvider):
    def __init__(self, config: AgentConfig) -> None:
        self.client = AsyncOpenAI(
            api_key=os.environ["RELACE_API_KEY"],
            base_url="https://search.endpoint.relace.run/v1/search",
            timeout=config.prompt_timeout,
            max_retries=config.prompt_retries,
        )

    @override
    async def generate_response(
        self,
        context: Context,
        config: AgentConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam],
    ) -> ChatCompletion:
        response = await self.client.chat.completions.create(
            prompt_cache_key=context.prompt_cache_key,  # results in sticky routing by key
            messages=messages,
            model=config.model_name,
            tools=tools,
            max_tokens=config.max_tokens,
        )
        if not response.choices:
            raise RuntimeError("Received empty choices")
        return response



import asyncio
import logging
import os
import uuid
from pathlib import Path

from fastmcp import Context as MCPContext
from fastmcp import FastMCP

from relace_agent.agents import Agent
from relace_agent.config import AgentConfig, LiteralPrompt
from relace_agent.context import ChatHistory, Context, EventStream
from relace_agent.server.types import AgentEvent, ToolEvent
from relace_toolbox.local_client import ToolboxClient
from relace_toolbox.tools import ReportBackToolSchema

REQUIRED_ENV_VARS = ["RELACE_API_KEY"]

missing = [var for var in REQUIRED_ENV_VARS if not os.environ.get(var)]
if missing:
    raise RuntimeError(
        f"Missing required environment variables: {', '.join(missing)}. "
        f"Set these in your MCP server config or shell environment."
    )

logging.basicConfig(
    level=logging.WARNING,  # Root logger at WARNING to suppress dependencies
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
# Set first-party loggers to DEBUG/INFO
for logger_name in [
    "relace_mcp",
    "relace_agent",
    "relace_toolbox",
]:
    logging.getLogger(logger_name).setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)

mcp = FastMCP("Relace MCP Server")

FAS_TOOLS = [
    "view_file",
    "view_directory",
    "grep_search",
    "bash",
    "report_back",
]

FAS_SYSTEM_PROMPT = """
You are an AI agent whose job is to explore a code base with the provided tools and thoroughly understand the problem.
You should use the tools provided to explore the codebase, read files, search for specific terms, and execute bash
commands as needed. Once you have a good understanding of the problem, use the `report_back` tool share your findings.
Make sure to only use the `report_back` tool when you are confident that you have gathered enough information to make
an informed decision. Your objective is speed and efficiency so call multiple tools at once where applicable to reduce
latency and reduce the number of turns. You are given a limited number of turns so aim to call 4-12 tools in parallel.
You are suggested to explain your reasoning for the tools you choose to call before calling them.
"""

FAS_USER_PROMPT = """
I have uploaded a code repository in the ${{ context.workdir }} directory.

Now consider the following user query:

<user_query>
${{ inputs.query }}
</user_query>

You need to resolve the <user_query>.

To do this, follow the workflow below:

---
Your job is purely to understand the codebase.
### 1. Explore and Understand the Codebase

You **must first build a deep understanding of the relevant code**.
Use the available tools to:

- Locate and examine all relevant parts of the codebase.
- Understand how the current code works, including expected behaviors, control flow, and edge cases.
- Identify the potential root cause(s) of the issue or the entry points for the requested feature.
- Review any related unit tests to understand expected behavior.

---

### 2. Report Back Your Understanding

Once you believe you have a solid understanding of the issue and the relevant code:

- Use the `report_back` tool to report you findings.
    - File paths should be relative to the working directory; failure to comply will result in deductions.
    - Only report the relevant files within the repository. You may speculate that a file or folder may be added in your explaination, but it must not be put within you reported files.

---

### Success Criteria

A successful resolution means:

- The specific issue in the <user_query> is well understood.
- Your explain clearly the reasoning behind marking code as relavent.
- The files comprehensively covers all the key files needed to address the query.
    - Relevant files can be any of three types:
    - Files needing edits
    - Files providing needed provide the required edits


<use_parallel_tool_calls>
If you intend to call multiple tools and there are no dependencies between the tool calls, make all of the independent tool calls in parallel. Prioritize calling tools simultaneously whenever the actions can be done in parallel rather than sequentially. For example, when reading 3 files, run 3 tool calls in parallel to read all 3 files into context at the same time. Maximize use of parallel tool calls where possible to increase speed and efficiency. However, if some tool calls depend on previous calls to inform dependent values like the parameters, do NOT call these tools in parallel and instead call them sequentially. Never use placeholders or guess missing parameters in tool calls.
Parallel tool calls can be made using the following schema:
<tool_call>
<function=example_function_name_1>
<parameter=example_parameter_1>
value_1\n</parameter>
<parameter=example_parameter_2>
</parameter>
</function>
<function=example_function_name_2>
<parameter=example_parameter_1>
value_1
</parameter>
<parameter=example_parameter_2>
</parameter>
</function>
</tool_call>
Where you can place as many <function=...>...</function> tags as you want within the <tool_call>...</tool_calls> tags for parallel tool calls.
</use_parallel_tool_calls>
</use_parallel_tool_calls>
"""


@mcp.tool
async def relace_search(
    query: str,
    workdir: str,
    ctx: MCPContext,
) -> dict[str, list[tuple[int, int]]]:
    """Agentic codebase search powered by Relace. Use when finding files to modify or answering questions requiring deep search.

    Args:
        query: Specific, human-readable search query (avoid minimal keywords).
            Examples:
            - "Find where user authentication tokens are validated"
            - "How does the payment processing flow work?"
            - "Where are database migrations defined and executed?"
        workdir: Absolute path to search directory.
    """

    workdir_path = Path(workdir)

    event_stream = EventStream()

    async def _run_agent() -> None:
        try:
            agent = Agent(
                config=AgentConfig(
                    name="relace-search",
                    model_name="relace-search",
                    prompts=LiteralPrompt(
                        system_prompt=FAS_SYSTEM_PROMPT,
                        user_prompt=FAS_USER_PROMPT,
                    ),
                    tools=FAS_TOOLS,
                ),
            )
            agent_context = Context(
                prompt_cache_key=f"fas-{uuid.uuid4()}",
                toolbox=ToolboxClient(),
                toolbox_workdir=workdir_path,
                inputs={
                    "query": query,
                },
                history=ChatHistory(max_messages=None),
                events=event_stream,
            )
            await agent.run_loop(agent_context)
        finally:
            await event_stream.close()

    agent_task = asyncio.create_task(_run_agent())

    # Process events concurrently with agent execution
    final_result: dict[str, list[tuple[int, int]]] | None = None
    try:
        tool_call_count = 0
        async for event in event_stream:
            logger.info("Event: %s", event)

            # Emit tool calls as progress notifications to make them visible in Cursor
            if isinstance(event, ToolEvent):
                tool_call_count += 1

                # Format tool call information for progress notification
                tool_info = f"🔧 {event.name}"
                if event.path:
                    tool_info += f": {event.path}"
                elif event.query:
                    query_preview = (
                        event.query[:50] + "..."
                        if len(event.query) > 50
                        else event.query
                    )
                    tool_info += f": {query_preview}"
                elif event.command:
                    cmd_preview = (
                        event.command[:50] + "..."
                        if len(event.command) > 50
                        else event.command
                    )
                    tool_info += f": {cmd_preview}"

                # Report progress to make tool calls visible in Cursor's chat
                await ctx.report_progress(
                    progress=tool_call_count,
                    message=tool_info,
                )

                logger.info("🔧 %s", tool_info)

                # Check if this is the final report_back event
                if event.name == ReportBackToolSchema.tool_name:
                    if not event.files:
                        raise RuntimeError("Missing files in report back event")
                    final_result = event.files
            elif isinstance(event, AgentEvent):
                # Show agent messages as progress
                content_preview = (
                    event.content[:100] + "..."
                    if len(event.content) > 100
                    else event.content
                )
                await ctx.report_progress(
                    progress=0,
                    message=f"💭 {content_preview}",
                )
                logger.info("💭 Agent: %s", content_preview)
    except Exception as e:
        logger.error("Error processing events: %s", e, exc_info=True)
        raise
    finally:
        if not agent_task.done():
            agent_task.cancel()

    if final_result is not None:
        return final_result

    raise RuntimeError("No report back event emitted")


def main() -> None:
    mcp.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import logging
import shlex
import subprocess
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Self, final, override

import aiofiles
from openai import pydantic_function_tool
from openai.types.chat import ChatCompletionToolParam
from pydantic import BaseModel, Field, ValidationError, field_validator

from relace_toolbox.errors import (
    ExitAgentLoop,
    InvalidToolInput,
    InvalidToolName,
    ToolExecutionError,
    ToolExecutionTimeout,
)
from relace_toolbox.types import ToolResult

logger = logging.getLogger(__name__)


class ToolSchema(BaseModel):
    """Base class for tool input parameters."""

    tool_name: ClassVar[str]
    tool_description: ClassVar[str]

    _tool_lookup: ClassVar[dict[str, type[ToolSchema]]] = {}

    def __init_subclass__(
        cls,
        name: str,
        description: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        description = description or cls.__doc__
        if not description:
            raise ValueError(f"{cls.__name__} is missing a docstring or description")
        cls.tool_name = name
        cls.tool_description = description.strip()
        cls._tool_lookup[name] = cls

    @classmethod
    def from_name(cls, tool_name: str) -> type[ToolSchema]:
        try:
            return cls._tool_lookup[tool_name]
        except KeyError as e:
            raise InvalidToolName(tool_name) from e

    @classmethod
    def openai(cls) -> ChatCompletionToolParam:
        schema = pydantic_function_tool(
            model=cls,
            name=cls.tool_name,
            description=cls.tool_description,
        )

        # Remove unsupported 'format' from schema for Path fields, recursively
        def remove_path_format(obj: Any) -> None:
            if isinstance(obj, dict):
                if obj.get("format") == "path":
                    obj.pop("format")
                for val in obj.values():
                    remove_path_format(val)
            elif isinstance(obj, list):
                for item in obj:
                    remove_path_format(item)

        schema_params = schema["function"]["parameters"]
        remove_path_format(schema_params)
        return schema

    @classmethod
    def from_json_str(cls, json_str: str) -> Self:
        try:
            return cls.model_validate_json(json_str)
        except ValidationError as e:
            raise InvalidToolInput(json_str) from e

    @classmethod
    def from_json_dict(cls, json_dict: dict[str, Any]) -> Self:
        try:
            return cls.model_validate(json_dict)
        except ValidationError as e:
            raise InvalidToolInput(json_dict) from e

    def event_data(self) -> dict[str, Any]:
        """Optional event data to include in event stream when calling this tool."""
        return {}


@dataclass
class ToolContext:
    workdir: Path
    chat_history: str | None


class Tool[S: ToolSchema](ABC):
    """Base class for tools that an agent can use."""

    schema: ClassVar[type[S]]
    requires_chat_history: ClassVar[bool] = False

    _tools: ClassVar[dict[str, type[Tool[ToolSchema]]]] = {}

    def __init_subclass__(
        cls,
        schema: type[S],
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        cls.schema = schema
        cls._tools[schema.tool_name] = cls  # type: ignore

    @abstractmethod
    async def execute(
        self,
        tool_input: S,
        context: ToolContext,
    ) -> ToolResult:
        """Execute the tool with the given parameters.

        Args:
            tool_input: Structured, validated input arguments for the tool.
        """
        ...

    @classmethod
    def from_name(cls, name: str) -> Tool[ToolSchema]:
        try:
            return cls._tools[name]()
        except KeyError as e:
            raise InvalidToolName(name) from e

    @classmethod
    def all_tools(cls) -> list[type[Tool[ToolSchema]]]:
        return list(cls._tools.values())


class BashToolSchema(ToolSchema, name="bash"):
    """Tool for executing bash commands.

    * Avoid long running commands
    * Avoid dangerous/destructive commands
    * Prefer using other more specialized tools where possible
    """

    command: str = Field(..., description="Bash command to execute")

    @override
    def event_data(self) -> dict[str, str]:
        return {"command": self.command}


@final
class BashTool(Tool[BashToolSchema], schema=BashToolSchema):
    timeout: ClassVar[int] = 3
    text_limit: ClassVar[int] = 5_000

    @override
    async def execute(
        self,
        tool_input: BashToolSchema,
        context: ToolContext,
    ) -> ToolResult:
        def truncate(text: str) -> str:
            if len(text) > self.text_limit:
                return (
                    f"{text[: self.text_limit]}\n"
                    f"...truncated to {self.text_limit} characters ..."
                )
            return text

        try:
            # Execute the bash command and capture output
            proc = await asyncio.create_subprocess_shell(
                # NOTE: We must run in a clean login shell to prevent the toolbox
                # venv from superceding system python and PYTHONPATH
                f"env -i bash -lc {shlex.quote(tool_input.command)}",
                cwd=context.workdir,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = await asyncio.wait_for(
                proc.communicate(),
                timeout=self.timeout,
            )
        except TimeoutError as e:
            raise ToolExecutionTimeout(
                f"Command timed out after {self.timeout} seconds."
            ) from e

        if proc.returncode != 0:
            raise ToolExecutionError(
                f"Command failed with exit code {proc.returncode}.\n"
                f"STDOUT:\n{truncate(stdout.decode())}\n"
                f"STDERR:\n{truncate(stderr.decode())}"
            )
        return ToolResult(message=truncate(stdout.decode()))


class DirectoryViewToolSchema(ToolSchema, name="view_directory"):
    """Tool for viewing the contents of a directory.

    * Lists contents recursively, relative to the input directory
    * Directories are suffixed with a trailing slash '/'
    * Depth might be limited by the tool implementation
    * Output is limited to the first 250 items

    Example output:
    file1.txt
    file2.txt
    subdir1/
    subdir1/file3.txt
    """

    path: Path = Field(
        ...,
        description="Absolute path to a directory, e.g. `/repo/`.",
    )
    include_hidden: bool = Field(
        False,
        description="If true, include hidden files in the output (false by default).",
    )

    @override
    def event_data(self) -> dict[str, str]:
        return {"path": str(self.path)}


class DirectoryViewTool(Tool[DirectoryViewToolSchema], schema=DirectoryViewToolSchema):
    limit: ClassVar[int] = 250

    @override
    async def execute(
        self,
        tool_input: DirectoryViewToolSchema,
        context: ToolContext,
    ) -> ToolResult:
        check_path(tool_input.path, context.workdir, must_exist=True, file_ok=False)

        i = 0
        contents = []
        for item in tool_input.path.rglob("*"):
            i += 1
            if i > self.limit:
                contents.append("... remaining items omitted ...")
                break

            if not tool_input.include_hidden and item.name.startswith("."):
                continue

            relative_path = (
                item.relative_to(context.workdir) if context.workdir else item
            )
            if item.is_dir():
                contents.append(f"{relative_path}/")
            else:
                contents.append(f"{relative_path}")

        return ToolResult(
            message="\n".join(contents) if contents else "Directory is empty."
        )


class FileViewToolSchema(ToolSchema, name="view_file"):
    """Tool for viewing/exploring the contents of existing files

    Line numbers are included in the output, indexing at 1. If the output does not
    include the end of the file, it will be noted after the final output line.

    Example (viewing the first 2 lines of a file):
    1\tdef my_function():
    2\t    print("Hello, World!")
    ... rest of file truncated ...
    """

    path: Path = Field(
        ...,
        description="Absolute path to a file, e.g. `/repo/file.py`.",
    )
    # TODO: Split this into two integers for more precise validation
    # NOTE: Tuple validation is not supported by the OpenAI API, so we cannot do that
    view_range: list[int] = Field(
        [1, 100],
        description="Range of file lines to view. If not specified, the first 100 lines "
        "of the file are shown. If provided, the file will be shown in the indicated "
        "line number range, e.g. [11, 12] will show lines 11 and 12. Indexing at 1 to "
        "start. Setting `[start_line, -1]` shows all lines from `start_line` to the end "
        "of the file.",
    )

    @override
    def event_data(self) -> dict[str, str]:
        return {"path": str(self.path)}


@final
class FileViewTool(Tool[FileViewToolSchema], schema=FileViewToolSchema):
    @override
    async def execute(
        self,
        tool_input: FileViewToolSchema,
        context: ToolContext,
    ) -> ToolResult:
        check_path(tool_input.path, context.workdir, must_exist=True, dir_ok=False)

        start, end = tool_input.view_range
        view_lines: list[str] = []
        async with aiofiles.open(tool_input.path) as handle:
            for i, line in enumerate(await handle.readlines(), start=1):
                if end != -1 and i > end:
                    view_lines.append("... rest of file truncated ...")
                    break
                if i >= start:
                    view_lines.append(f"{i}\t{line}")

        return ToolResult(message="".join(view_lines))


class RipgrepToolSchema(ToolSchema, name="grep_search"):
    """Fast text-based regex search that finds exact pattern matches within files or
    directories, utilizing the ripgrep command for efficient searching. Results will be
    formatted in the style of ripgrep and can be configured to include line numbers and
    content. To avoid overwhelming output, the results are capped at 50 matches. Use the
    include or exclude patterns to filter the search scope by file type or specific
    paths. This is best for finding exact text matches or regex patterns. More precise
    than semantic search for finding specific strings or patterns. This is preferred
    over semantic search when we know the exact symbol/function name/etc. to search in
    some set of directories/file types.
    """

    query: str = Field(..., description="The regex pattern to search for")
    case_sensitive: bool = Field(
        default=True, description="Whether the search should be case sensitive"
    )
    exclude_pattern: str | None = Field(
        default=None, description="Glob pattern for files to exclude"
    )
    include_pattern: str | None = Field(
        default=None,
        description="Glob pattern for files to include (e.g. '.ts' for TypeScript files)",
    )

    @override
    def event_data(self) -> dict[str, str]:
        return {"query": self.query}


class RipgrepTool(Tool[RipgrepToolSchema], schema=RipgrepToolSchema):
    @override
    async def execute(
        self,
        tool_input: RipgrepToolSchema,
        context: ToolContext,
    ) -> ToolResult:
        args = ["rg", tool_input.query, "./", "--max-count=50", "--color=never"]
        if not tool_input.case_sensitive:
            args.append("--ignore-case")
        if tool_input.exclude_pattern:
            args.append(f"--glob=!{tool_input.exclude_pattern}")
        if tool_input.include_pattern:
            args.append(f"--glob={tool_input.include_pattern}")

        process = await asyncio.create_subprocess_exec(
            *args,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=context.workdir,
        )
        stdout, stderr = await process.communicate()
        if process.returncode == 0:
            return ToolResult(message=stdout.decode())
        elif process.returncode == 1:
            return ToolResult(message="No matches found")
        else:
            raise ToolExecutionError(f"Error running ripgrep: {stderr.decode()}")


class ReportBackToolSchema(ToolSchema, name="report_back"):
    """This tool is to be used to terminate the run when you have finished exploring
    the codebase and understand the problem.
    """

    explanation: str = Field(
        ...,
        description="Details your reasoning for deeming the files relevant for solving "
        "the issue.",
    )
    files: dict[str, list[tuple[int, int]]] = Field(
        ...,
        description="Key is a relevant file within the repo. The value is the ranges "
        "of relevant lines within the file.",
    )

    @field_validator("files", mode="before")
    @classmethod
    def validate_files(cls, data: Any) -> Any:
        if isinstance(data, dict):
            output: dict[Any, Any] = {}
            # Handle common invalid inputs for file ranges
            for k, v in data.items():
                if v is None:
                    output[k] = []  # None -> []
                elif (
                    v
                    and isinstance(v, list | tuple)
                    and all(isinstance(i, int) for i in v)
                ):
                    output[k] = [v]  # [1, 2] -> [[1, 2]]
                else:
                    output[k] = v
            return output
        else:
            return data

    @override
    def event_data(self) -> dict[str, Any]:
        return {
            "explanation": self.explanation,
            "files": self.files,
        }


class ReportBackTool(Tool[ReportBackToolSchema], schema=ReportBackToolSchema):
    @override
    async def execute(
        self,
        tool_input: ReportBackToolSchema,
        context: ToolContext,
    ) -> ToolResult:
        raise ExitAgentLoop("Report back to user")


def check_path(
    path: Path,
    workdir: Path,
    must_exist: bool = False,
    must_not_exist: bool = False,
    file_ok: bool = True,
    dir_ok: bool = True,
) -> None:
    if not path.is_relative_to(workdir):
        raise InvalidToolInput(
            f"path is not contained in the working directory: {path}"
        )
    if must_exist and not path.exists():
        raise InvalidToolInput(f"path does not exist: {path}")
    if must_not_exist and path.exists():
        raise InvalidToolInput(f"path already exists: {path}")
    if not file_ok and path.is_file():
        raise InvalidToolInput(f"path is a file: {path}")
    if not dir_ok and path.is_dir():
        raise InvalidToolInput(f"path is a directory: {path}")



from __future__ import annotations

from typing import ClassVar

from pydantic import (
    BaseModel,
)
from sse_starlette import ServerSentEvent


class BaseEvent(BaseModel):
    event_type: ClassVar[str | None] = None

    def to_sse(self) -> ServerSentEvent:
        return ServerSentEvent(
            event=self.event_type,
            data=self.model_dump_json(exclude_none=True),
        )


class AgentEvent(BaseEvent):
    event_type: ClassVar[str | None] = "agent"

    name: str
    content: str


class ToolEvent(BaseEvent):
    event_type: ClassVar[str | None] = "tool"

    name: str
    path: str | None = None
    query: str | None = None
    command: str | None = None

    # Report back tool
    explanation: str | None = None
    files: dict[str, list[tuple[int, int]]] | None = None



import logging
from pathlib import Path

from relace_toolbox.errors import (
    InvalidToolName,
    ToolError,
    ToolExecutionError,
    ToolExecutionTimeout,
)
from relace_toolbox.tools import Tool, ToolContext
from relace_toolbox.types import ToolResult, ToolSpec

logger = logging.getLogger(__name__)


class ToolboxClient:
    async def tool_specs(self) -> dict[str, ToolSpec]:
        return {
            tool.schema.tool_name: ToolSpec(
                tool_name=tool.schema.tool_name,
                tool_schema=tool.schema.openai(),
                requires_chat_history=tool.requires_chat_history,
            )
            for tool in Tool.all_tools()
        }

    async def tool_spec(self, tool_name: str) -> ToolSpec:
        tool_specs = await self.tool_specs()
        try:
            return tool_specs[tool_name]
        except KeyError as e:
            raise InvalidToolName(tool_name) from e

    async def tool_call(
        self,
        tool_name: str,
        tool_input: str,
        workdir: Path,
        chat_history: str | None = None,
    ) -> ToolResult:
        tool = Tool.from_name(tool_name)
        try:
            tool_result = await tool.execute(
                tool_input=tool.schema.from_json_str(tool_input),
                context=ToolContext(
                    workdir=workdir,
                    chat_history=chat_history,
                ),
            )
            return tool_result
        except ToolError as exc:
            logger.warning("Explicit tool error", exc_info=True)
            raise exc
        except TimeoutError as exc:
            raise ToolExecutionTimeout("Tool execution timed out") from exc
        except Exception as exc:
            logger.warning(f"Unexpected tool error: {exc}", exc_info=True)
            raise ToolExecutionError(f"{exc.__class__.__name__}: {exc!s}") from exc



class AgentStop(BaseException):
    """Control flow mechanism to break out of an iterator."""


class AgentError(Exception):
    """Error during agent's operation."""


class BuildError(AgentError):
    """Error during site build."""


class TestError(AgentError):
    """Error during testing of generated site."""



from __future__ import annotations

import asyncio
import logging
import re
from abc import ABCMeta, abstractmethod
from collections import deque
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar, Literal, override

from openai.types import CompletionUsage
from openai.types.chat import (
    ChatCompletionMessage,
    ChatCompletionMessageParam,
)
from pydantic import BaseModel

from relace_agent.server.types import (
    BaseEvent,
)
from relace_toolbox.local_client import ToolboxClient
from relace_toolbox.types import ToolResult

logger = logging.getLogger(__name__)


@dataclass
class ShellOutput:
    exit_code: int
    stdout: str
    stderr: str


class BaseChatItem(BaseModel, metaclass=ABCMeta):
    @abstractmethod
    def as_message_params(self) -> list[ChatCompletionMessageParam]: ...


class UserMessage(BaseChatItem):
    role: Literal["user"] = "user"
    content: str

    @override
    def as_message_params(self) -> list[ChatCompletionMessageParam]:
        return [
            {
                "role": self.role,
                "content": [
                    {
                        "type": "text",
                        "text": self.content,
                    }
                ],
            }
        ]


class AssistantMessage(BaseChatItem):
    role: Literal["assistant"] = "assistant"
    content: str

    @override
    def as_message_params(self) -> list[ChatCompletionMessageParam]:
        return [
            {
                "role": self.role,
                "content": self.content,
            }
        ]


class ToolUse(BaseChatItem):
    message: ChatCompletionMessage
    results: dict[str, ToolResult]

    @override
    def as_message_params(self) -> list[ChatCompletionMessageParam]:
        messages: list[ChatCompletionMessageParam] = [
            self.message.model_dump(mode="json", exclude_none=True)  # type: ignore
        ]
        for tool_call_id, result in self.results.items():
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call_id,
                    "content": [
                        {
                            "type": "text",
                            "text": result.message,
                        }
                    ],
                }
            )
            if result.image_url is not None:
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "image_url",
                                "image_url": {"url": result.image_url},
                            }
                        ],
                    }
                )
        return messages


# Discriminated union needed for serialization to work correctly
ChatItem = UserMessage | AssistantMessage | ToolUse


# TODO: Intelligent file context (one copy at a time, not written to disk)
# TODO: Intelligent truncation/summarization of history
# TODO: Encapsulate tool definitions and system prompt
class ChatHistory:
    def __init__(
        self,
        max_messages: int | None = 100,
        max_tokens: int = 140_000,
    ) -> None:
        self.messages: deque[ChatItem] = deque(maxlen=max_messages)
        self.max_tokens: int = max_tokens
        self.last_tokens: int = 0

    def append(self, item: ChatItem) -> None:
        self.messages.append(item)

    def as_params(self) -> list[ChatCompletionMessageParam]:
        return [
            param for message in self.messages for param in message.as_message_params()
        ]

    def record_usage(
        self, usage: CompletionUsage | None, auto_clean: bool = True
    ) -> None:
        if usage is not None:
            self.last_tokens = usage.total_tokens
        # Truncate history if we're over our limit
        if auto_clean and self.last_tokens > self.max_tokens:
            self.clean()

    # TODO (eborgnia): This is the simplest solution, but we should consider more sophisticated approaches.
    # - Remove unecessary file reads that pollute the context window.
    # - Delete the center messages instead to not lose the original user intent.
    # - Maintain only summary messages or some kind of markdown log of the changes.
    def clean(self) -> None:
        """
        Simple conversation cleaning by removing the oldest non-user message.
        This helps manage token limits by gradually removing old context.
        """
        logger.info("Cleaning conversation history")
        messages_list = list(self.messages)
        for i, message in enumerate(messages_list[:]):
            if not isinstance(message, UserMessage):
                del messages_list[i]
                self.messages.clear()
                self.messages.extend(messages_list)
                logger.info("Removed message: %s", message)
                break
        else:
            logger.info("No messages to remove from conversation history")

    def tokenize(self) -> str:
        chunks: list[str] = []
        for message in self.messages:
            if isinstance(message, UserMessage):
                chunks.append(f"<user>{message.content}</user>")
            elif isinstance(message, AssistantMessage):
                chunks.append(f"<assistant>{message.content}</assistant>")
            elif isinstance(message, ToolUse) and message.message.tool_calls:
                chunks.extend(
                    f"<tool_call>{tool_call.function.name}</tool_call>"  # type: ignore
                    for tool_call in message.message.tool_calls
                )
        return "\n".join(chunks)


class EventStream:
    def __init__(self) -> None:
        self._queue: asyncio.Queue[BaseEvent | None] = asyncio.Queue()
        self._closed: bool = False

    async def emit(self, event: BaseEvent) -> None:
        if self._closed:
            raise RuntimeError(f"{self.__class__.__name__} is closed")

        await self._queue.put(event)

    async def close(
        self,
        drain: bool = True,
    ) -> None:
        self._closed = True
        self._queue.put_nowait(None)

        if drain:
            await self._queue.join()

    async def __aiter__(self) -> AsyncIterator[BaseEvent]:
        while True:
            event = await self._queue.get()
            try:
                if event is None:
                    break
                yield event
            finally:
                self._queue.task_done()


@dataclass
class Context:
    prompt_cache_key: str
    toolbox: ToolboxClient
    toolbox_workdir: Path = field(default_factory=Path.cwd)
    inputs: dict[str, str] = field(default_factory=dict)
    history: ChatHistory = field(default_factory=ChatHistory)
    events: EventStream = field(default_factory=EventStream)

    _interpolate_pattern: ClassVar[re.Pattern[str]] = re.compile(
        r"\$\{\{ (?P<namespace>\w+)(\.(?P<key>\w+))? \}\}"
    )

    def interpolate(self, template: str) -> str:
        def handle_match(match: re.Match[str]) -> str:
            namespace = match.group("namespace")
            key = match.group("key")
            match (namespace, key):
                case ("context", "workdir"):
                    return self.toolbox_workdir.as_posix()
                case ("inputs", input_key):
                    return self.inputs[input_key]
                case _:
                    raise ValueError(f"Invalid substitution: {namespace}.{key}")

        return self._interpolate_pattern.sub(handle_match, template)



from pathlib import Path
from typing import ClassVar, Self, TypedDict

from pydantic import BaseModel, ConfigDict
from pydantic_yaml import parse_yaml_file_as

CONFIG_ROOT = Path(__file__).parent.parent.parent / "config"
EVAL_CONFIG_ROOT = CONFIG_ROOT / "evals"
API_CONFIG_ROOT = CONFIG_ROOT / "api"


class ChatMessage(TypedDict):
    role: str
    content: str


class BaseConfig(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")


class FileConfig(BaseConfig):
    @classmethod
    def load(cls, path: Path) -> Self:
        return parse_yaml_file_as(cls, path)


class LiteralPrompt(BaseModel):
    system_prompt: str
    user_prompt: str


class AgentConfigOverrides(BaseModel):
    prompts: LiteralPrompt
    tools: list[str] | None = None
    model_name: str | None = None
    max_tokens: int | None = None
    max_turns: int | None = None
    prompt_timeout: int | None = None  # seconds
    prompt_retries: int | None = None


class AgentConfig(BaseConfig):
    name: str
    prompts: LiteralPrompt
    tools: list[str]

    model_name: str = "relace-search"
    max_tokens: int = 8192
    max_turns: int = 200
    prompt_timeout: int = 180  # 3 minutes
    prompt_retries: int = 1

    def with_overrides(self, overrides: AgentConfigOverrides) -> Self:
        return self.model_validate(
            self.model_dump() | overrides.model_dump(exclude_none=True),
            strict=True,
        )



import base64
from typing import Literal, Self

from openai.types.chat import ChatCompletionToolParam
from pydantic import BaseModel


class ToolResult(BaseModel):
    message: str
    image_url: str | None = None

    @classmethod
    def from_image_bytes(
        cls,
        message: str,
        image: bytes,
        image_format: Literal["image/jpeg", "image/png", "image/webp", "image/gif"],
    ) -> Self:
        return cls(
            message=message,
            image_url=f"data:{image_format};base64,{base64.b64encode(image).decode()}",
        )


class ToolSpec(BaseModel):
    tool_name: str
    tool_schema: ChatCompletionToolParam
    requires_chat_history: bool



from __future__ import annotations

from typing import ClassVar


class ToolError(Exception):
    """Base class for custom errors in relace-toolbox."""

    error_code: ClassVar[str] = "unknown_tool_error"

    _error_lookup: dict[str, type[ToolError]] = {}

    def __init_subclass__(cls) -> None:
        if cls.error_code in cls._error_lookup:
            raise ValueError(f"Duplicate error code: {cls.error_code}")
        cls._error_lookup[cls.error_code] = cls
        return super().__init_subclass__()

    def to_json(self) -> dict[str, str]:
        return {
            "error_code": self.error_code,
            "message": str(self),
        }

    @classmethod
    def from_json(cls, data: dict[str, str]) -> ToolError:
        error_code = data.get("error_code", "unknown_tool_error")
        error_class = cls._error_lookup.get(error_code, ToolError)
        message = data.get("message", "")
        return error_class(message)


class InvalidToolName(ToolError):
    """The provided tool name is not recognized."""

    error_code = "invalid_tool_name"


class InvalidToolInput(ToolError):
    """The provided tool input does not conform to the tool's schema, or is otherwise invalid/forbidden."""

    error_code = "invalid_tool_input"


class ToolExecutionError(ToolError):
    """An error occurred while the tool was executing."""

    error_code = "tool_execution_error"


class ToolExecutionTimeout(ToolError):
    """The tool execution, or one of its dependencies, exceeded the allowed time limit."""

    error_code = "tool_execution_timeout"


class ExitAgentLoop(ToolError):
    """Signal for the agent to exit its loop."""

    error_code = "exit_agent_loop"
